@@ -321,11 +321,10 @@ def cosine_kernel(x, y):
 kernel_fullnames = {
     "poly": "polynomial",
     "rbf": "radial basis function",
-    "laplace": "laplacian",
+    "laplace": "laplacian / exponential",
     "cosine": "cosine similarity",
     "chi2": "chi-squared",
     "achi2": "additive chi-squared",
-    "exp": "exponential",
     "ess": "exponential sine squared",
     "rq": "rational quadratic",
     "imq": "inverse multi quadric",
@@ -334,6 +333,9 @@ kernel_fullnames = {
     "minman": "min-max",
     "expmin": "exponential-min",
     "ghi": "generalized histogram intersection",
+    "cauchy": "Cauchy",
+    "fejer": "Fejér",
+    "matern": "Matérn",
 }
 
 # kernel names and functions
@@ -341,7 +343,7 @@ kernel_dict = {"linear": linear_kernel,
                "poly": polynomial_kernel,
                "sigmoid": sigmoid_kernel,
                "gaussian": gaussian_kernel,  # alias of "rbf": rbf_kernel,
-               "exp": laplacian_kernel,  # alias of laplacian
+               # "exp": laplacian_kernel,  # alias of laplacian
                "laplace": laplacian_kernel,
                "matern": matern_kernel,
                "chi2": chi2_kernel,
@@ -380,7 +382,7 @@ kernel_formulas = {
     "exp": r"$k(x, y)=exp(-||x - y||/(2 \sigma^2))$",
     "laplace": r"$k(x, y) = exp(-||x - y||/ \sigma)$",
     "cosine": r"$k(x, y) = <x,y>/(||x|| ||y||)$",
-    "wave": r"$k(x, y) = \frac{\sigma}{\lVert x-y \rVert } \sin \frac{\lVert x-y \rVert }{\sigma}$",
+    "wave": r"$k(x, y) = \frac{\sigma}{|x-y|} \sin \frac{|x-y|}{\sigma}$",
     "matern": r"$k(x, y) = (||x-y||^v \sqrt{2v} / s) * Bessel(||x-y||^v \sqrt{2v} / s) /(\Gamma(v) 2^{v-1} ) $",
     "rq": r"$k(x, y) = \left(1 + \frac{ {||x - y||} ^2 }{ 2\alpha  l^2}\right)^{-\alpha} $",
     "imq": r"$k(x, y) = 1 / \sqrt{||x-y||^2 + c^2}$",
@@ -389,8 +391,8 @@ kernel_formulas = {
     "anova": r"$k(x, y) = \sum_i exp( - \sigma * (x_i - y_i)^2 )^d$",
     "wavelet": r"$k(x, y) = \prod_i { h( (x_i-y_i)/a )}$",
     "fourier": r"$k(x, y) = \prod_i { (1-q^2)/(2(1-2q cos(||x - y||)+q^2)) }$",
-    "circular": r"$k(x, y) = \frac{2}{\pi} \arccos ( \frac{ \lVert x-y \rVert}{\sigma}) - \frac{2}{\pi} \frac{ \lVert x-y \rVert}{\sigma} \sqrt{1 - {\left( \frac{ \lVert x-y \rVert}{\sigma} \right)}^2 } , if \lVert x-y \rVert < \sigma$ , 0 otherwise.",
-    "spherical": r"$k(x, y) = 1 - \frac{3}{2} \frac{\lVert x-y \rVert}{\sigma} + \frac{1}{2} \left( \frac{ \lVert x-y \rVert}{\sigma} \right)^3$",
+    "circular": r"$k(x, y) = \frac{2}{\pi} \arccos ( \frac{ | x-y | }{\sigma}) - \frac{2}{\pi} \frac{ | x-y | }{\sigma} \sqrt{1 - {\left( \frac{ | x-y |}{\sigma} \right)}^2 } , if | x-y | < \sigma$ , 0 otherwise.",
+    "spherical": r"$k(x, y) = 1 - \frac{3}{2} \frac{| x-y |}{\sigma} + \frac{1}{2} \left( \frac{ | x-y |}{\sigma} \right)^3$",
     "tanimoto": r"$k(x, y) = <x, y> / (||x||^2 + ||y||^2 - <x, y>)$",
     "sorensen": r"$k(x, y) = 2 <x, y> / (||x||^2 + ||y||^2)$",
     "achi2": r"$k(x, y) = \sum_i { 2 x_i y_i / (x_i + y_i) }$",
@@ -412,24 +414,24 @@ kernel_formulas = {
 # Some hparams are dynamic (based on data dim).
 # Not all kernels have tunable hyper-parameters.
 kernel_hparams = {
-    "poly": [1, 2, 3, 4],  # when d = 1, becomes a linear kernel
-    "gaussian": [1, 3.33, 10, 60, 100, 300, 400],
-    "sigmoid": [0.1, 1, 10],
-    "laplace": [0.1, 1, 10, 30, 50],
-    "exp": [0.01, 0.03, 0.05, 0.1],
-    "chi2":  [0.0001, 0.001, 0.01, 0.1, 1.0, 10, 100],
-    "anova": [0.00001, 0.001, 0.1, 1, 10, 1000, 100000],
-    "cauchy": [1, 10, 100, 1000, 10000],
-    "power": [.0001, .001, .01, .1, 0.25, .5, 1],
-    "matern": [0.1, 0.5, 1, 10],
-    "ess": [0.01, 0.1, 0.5, 1, 2, 10],
+    "poly": [2, 3, 4],  # skip d = 1, as it degrades to a linear kernel
+    "gaussian": [0.1, 1, 3.33, 10, 60, 100, 300],
+    "sigmoid": [0.01, 0.1, 1, 10],
+    "laplace": [0.001, 0.01, 0.1, 1, 10, 30, 50],
+    # "exp": alias of laplace
+    "chi2":  [0.00001, 0.0001, 0.001, 0.01, 0.1, 1.0, 10, 100],
+    "anova": [0.000001, 0.00001, 0.001, 0.1, 1, 10, 1000, 100000],
+    "cauchy": [0.01, 0.1, 1, 10, 100, 1000, 10000],
+    "power": [.00001, .0001, .001, .01, .1, 1],
+    "matern": [0.01, 0.1, 0.5, 1, 10],
+    "ess": [0.001, 0.01, 0.1, 0.5, 1, 2, 10],
     "fejer": [2, 3, 4, 5],
-    "circular": [0.01, 0.1, 1, 10, 1000, 10000],
-    "spherical": [.001, 0.1, 10, 1000, 10000],
-    "wave": [1, 2, 10, 100, 1000, 10000],
-    "fourier": [0.05, 0.1, 0.2, 0.4, 0.8, 2],
-    "rq": [0.5, 1, 2, 4, 16],
-    "imq": [0.5, 1, 2],
+    "circular": [0.001, 0.01, 0.1, 1, 10, 1000, 10000],
+    "spherical": [.0001, .001, 0.1, 10, 1000, 10000],
+    "wave": [0.1, 1, 2, 10, 100, 1000, 10000],
+    "fourier": [0.001, 0.01, 0.05, 0.1, 0.2, 0.4, 0.8, 2],
+    "rq": [0.1, 0.5, 1, 2, 4, 16],
+    "imq": [0.1, 0.5, 1, 2],
 }
 
 kernel_hparas_divide_n = ['gaussian', 'sigmoid', 'laplace']
@@ -796,7 +798,10 @@ class ANOVA(Kernel):
 
 
 def default_wavelet(x):
-    return np.cos(1.75*x)*np.exp(-x**2/2)
+    '''
+    Default wavelet is the morlet wavelet.
+    '''
+    return np.cos(5*x)*np.exp(-x**2/2) # 1.75
 
 
 class Wavelet(Kernel):
