@@ -1,22 +1,36 @@
+import os
+import sys
 import math
 import timeit
+from itertools import combinations
 import numpy as np
+import pickle
 import matplotlib.pyplot as plt
-# import full name to avoid conflict with those display params
-import IPython.core.display
+import IPython.display
 
-from sklearn.linear_model import LogisticRegressionCV
-from sklearn.preprocessing import MinMaxScaler
 from sklearn.decomposition import PCA
 from sklearn.cross_decomposition import PLSRegression
 from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
-from cla.metrics import get_metrics, metric_polarity_dict, es_max
-from .kernels import kernel_names, kernel_hparams, kernel_formulas, \
+from sklearn.linear_model import LogisticRegressionCV
+from sklearn.preprocessing import MinMaxScaler
+from sklearn.decomposition import PCA
+from sklearn.model_selection import train_test_split
+from cla.metrics import get_metrics, metric_polarity_dict, es_max, run_multiclass_clfs
+from cla.vis.plt2base64 import plt2html
+
+if __package__:
+    from .kernels import kernel_names, kernel_hparams, kernel_formulas, \
+    kernel_fullnames, kernel_dict, kernel_hparas_divide_n, \
+    cosine_kernel
+else:
+    ROOT_DIR = os.path.dirname(__file__)
+    if ROOT_DIR not in sys.path:
+        sys.path.append(ROOT_DIR)
+    from kernels import kernel_names, kernel_hparams, kernel_formulas, \
     kernel_fullnames, kernel_dict, kernel_hparas_divide_n, \
     cosine_kernel
 
-
-def acc(X, y, f):
+def acc(X, y, f, verbose = True):
     '''
     Use classification accuracy as a metric for kernel performance
 
@@ -32,14 +46,14 @@ def acc(X, y, f):
     XK = np.nan_to_num(XK, copy=False, nan=0)  # replace NaN with 0
 
     try:
-        clf = LogisticRegressionCV().fit(XK, y)
+        clf = LogisticRegressionCV(max_iter=1000, multi_class='multinomial').fit(XK, y)
         return clf.score(XK, y)
     except Exception as e:
-        print(e)
+        if verbose:
+            print(e)
         return np.NaN
 
-
-def kes(X, y, f):
+def kes(X, y, f, verbose = True):
     '''
     kernel effect size
 
@@ -59,22 +73,25 @@ def kes(X, y, f):
     try:
         lda = LinearDiscriminantAnalysis()
         X_dr = lda.fit_transform(XK, y)
-    except:    
-        print('Exception in LDA. Try PLS.')
+    except:
+        if verbose:
+            print('Exception in LDA. Try PLS.')
         try:
             pls = PLSRegression(n_components=len(XK), scale=False)
             X_dr = pls.fit(XK, y).transform(XK)
         except:
-            print('Exception in PLS. Use PCA instead.')
+            if verbose:
+                print('Exception in PLS. Use PCA instead.')
             try:
                 X_dr = PCA().fit_transform(XK)
             except:
                 X_dr = XK  # do without DR
 
     X_dr = np.nan_to_num(X_dr)
-    return es_max(X_dr, y)
+    return es_max(X_dr, y, verbose=verbose)
 
-def nmd(X, y, f, display=False):
+
+def nmd(X, y, f, display=False, verbose = True):
     '''
     Computes the normalized mean difference between the between-class and within-class matrices.
 
@@ -90,7 +107,11 @@ def nmd(X, y, f, display=False):
     '''
 
     labels = list(set(y))
-    assert len(labels) == 2
+    if len(labels) != 2:
+        if verbose:
+            print('NMD only supports binary classification. Return 0 by default.')
+        return 0
+    
     X1 = X[y == labels[0]]
     X2 = X[y == labels[1]]
 
@@ -151,8 +172,8 @@ def binary_response_pattern(cmap='gray'):
     cmap : color map scheme
     '''
     X = np.array([0, 1]).reshape(-1, 1)
-    _ = preview_kernels(X, np.array(
-        [0, 1]), cmap, None, True, False, False, False, False)
+    _ = classify_with_kernels(X, np.array(
+        [0, 1]), cmap, None, scale=True, clfs=[], do_cla=False)
 
 
 def linear_response_pattern(n=10, dim=1, cmap='gray'):
@@ -178,14 +199,19 @@ def linear_response_pattern(n=10, dim=1, cmap='gray'):
         X = np.vstack((np.hstack((X, zeros)), np.hstack((zeros, X))))
         y = [0] * n + [1] * n
 
-    _ = preview_kernels(X, np.array(y), cmap, None,
-                        True, False, False, False, False)
+    _ = classify_with_kernels(X, np.array(y), cmap, hyper_param_optimizer = None,
+                        scale = True, clfs = [], do_cla = False,
+                        plots = True, embed_title = False, verbose = False)
 
 
-def preview_kernels(X, y, cmap=None, hyper_param_optimizer=kes,
-                    scale=False, metrics=True, logplot=False, scatterplot=True, embed_title=True):
+def classify_with_kernels(X, y, cmap=None, hyper_param_optimizer=kes,
+                    scale=False, clfs=['LinearDiscriminantAnalysis()'], do_cla = False,
+                    multi_kernels=[1], multi_kernel_topk=-1,
+                    logplot=False, plots=True, embed_title=False,
+                    output_html=False,
+                    selected_kernel_names=None, verbose = True):
     '''
-    Try various kernel types to evaluate the pattern after kernel-ops.  
+    Try various kernel types for classification task.  
     Each kernel uses their own default/empirical paramters.    
     In binary classification, because the first and last half samples belong to two classes respectively. 
     We want to get a contrasive pattern, i.e., (LT、RB) and （LB、RT）have blocks of different colors. 
@@ -200,18 +226,37 @@ def preview_kernels(X, y, cmap=None, hyper_param_optimizer=kes,
         For small dataset, we recommend 'gray'. For complex dataset, we recommend 'viridis'.
     hyper_param_optimizer : which optimizer to optimize the hyper parameters for each kernel. 
         For real-world dataset, use kes by default. For toy dataset, set None to disable optimizer.
+        For multi-class dataset (y has more than 5 classes), use acc.
     scale : whether do feature scaling
-    metrics : whether calculate clam metrics.
+    clfs : a list of multi-class classfiers. pass 'all' to use all available classifiers.
+    multi_kernels : how many kernels to use. Default is [1].
+    multi_kernel_topk : the k best single kernels to use in multi-kernel combinations. Default is 5.
     logplot : whether to output the log-scale plot in parallel.
-    scatterplot : whether to ouptut the scatter plots after PCA / PLS, to check classifiability.
+    plots : whether to ouptut the kernel heatmaps and the scatter plots after PCA / PLS, to check classifiability.
         The PLS tries to maximize the covariance between X and Y.
     embed_title : whether embed the title in the plots. If not, will generate the title in HTML.
+    do_cla : whether to do classifiability analysis using the cla package.
+    selected_kernel_names : a list of kernel names to be process. 
+        If None or 'all', will use all kernels.
     '''
 
+    dic_test_accs = {}
     all_dic_metrics = {}
+    html_str = ''
 
     if scale:
         X = MinMaxScaler().fit_transform(X)
+        print('perform min-max scaling')
+
+    if clfs == 'all' or len(clfs) > 0:
+        result_html = '<h3>0. no kernel</h3><p>Classification on original dataset</p>'
+        dic, mc_html = run_multiclass_clfs(X, y, clfs=clfs, show = False)
+        dic_test_accs['no kernel'] = dic
+        result_html += mc_html
+        if output_html:
+            html_str += result_html
+        if plots:
+            IPython.display.display(IPython.display.HTML(result_html))
 
     # to be safe, perform a re-order
     if (y is not None and len(set(y)) == 2):
@@ -220,7 +265,12 @@ def preview_kernels(X, y, cmap=None, hyper_param_optimizer=kes,
         y = [labels[0]] * np.sum(y == labels[0]) + \
             [labels[1]] * np.sum(y == labels[1])
 
-    for i, key in enumerate(kernel_names):
+    if selected_kernel_names is None or selected_kernel_names == 'all':
+        selected_kernel_names = kernel_names
+
+    KX = {} # store the data matrix after each kernel
+
+    for i, key in enumerate(selected_kernel_names):
         best_hparam = None
         best_metric = -np.inf
         title = str(i+1) + '. ' + \
@@ -231,7 +281,7 @@ def preview_kernels(X, y, cmap=None, hyper_param_optimizer=kes,
                 if key in kernel_hparas_divide_n:
                     param = param/X.shape[1]  # divide hparam by data dim
                 new_metric = hyper_param_optimizer(
-                    X, y, lambda x, y: kernel_dict[key](x, y, param))
+                    X, y, lambda x, y: kernel_dict[key](x, y, param), verbose = verbose)
                 if (new_metric > best_metric):
                     best_metric = new_metric
                     best_hparam = param
@@ -241,114 +291,231 @@ def preview_kernels(X, y, cmap=None, hyper_param_optimizer=kes,
 
         if hyper_param_optimizer is None or key not in kernel_hparams:
             kns = kernel_dict[key](X, X)
-            metric_nmd = nmd(X, y, lambda x, y: kernel_dict[key](x, y))
+            # metric_nmd = nmd(X, y, lambda x, y: kernel_dict[key](x, y), verbose = verbose)
         else:
             kns = kernel_dict[key](X, X, best_hparam)
-            metric_nmd = nmd(X, y, lambda x, y: kernel_dict[key](x, y, param))
+            # metric_nmd = nmd(X, y, lambda x, y: kernel_dict[key](x, y, param), verbose = verbose)
+
+        kns = np.nan_to_num(kns)
+        KX[key] = kns
 
         ######## plot after kernel transforms ########
-        _, ax = plt.subplots(1, 2, figsize=(12, 6))
-        ax[0].imshow(kns, cmap=cmap)
-        ax[0].set_axis_off()
-        if logplot:
-            ax[1].imshow(1+np.log(kns), cmap=cmap)
-            ax[1].set_axis_off()
+        
+        if plots or output_html:
+            _, ax = plt.subplots(1, 2, figsize=(round(len(set(y))/2.0) + 6, round(len(set(y))/4.0) + 3)) # figsize = (round(len(labels)/4.0) + 4, round(len(labels)/4.0) + 3)
+            ax[0].imshow(kns, cmap=cmap)
+            ax[0].set_axis_off()
+            if logplot:
+                ax[1].imshow(1+np.log(kns), cmap=cmap)
+                ax[1].set_axis_off()
 
-        plt.axis('off')
-        if embed_title:
-            plt.title(title + '\n' +
-                      kernel_formulas[key] + '\n' + "NMD = %.3g" % metric_nmd)
-        else:
-            # print(title)
-            IPython.display.display(IPython.display.HTML('<h3>' + title + '</h3>' + '<p>' + kernel_formulas[key].replace('<', '&lt;')
-                                                         .replace('>', '&gt;') + '</p><p>' + "NMD = %.3g" % metric_nmd + '</p>'))
-        plt.show()
+            plt.axis('off')
+            if embed_title:
+                plt.title(title + '\n' + kernel_formulas[key] + '\n') # + "NMD = %.3g" % metric_nmd)
+            else:
+                # print(title)
+                kernel_title = '<h3>' + title + '</h3>' + '<p>' + kernel_formulas[key].replace('<', '&lt;').replace('>', '&gt;') + '</p>'
+                if output_html:
+                    html_str += kernel_title
+                else:
+                    IPython.display.display(IPython.display.HTML(kernel_title)) # <p>' + "NMD = %.3g" % metric_nmd + '</p>'
+            if output_html:
+                html_str += plt2html(plt)
+                plt.close()
+            else:
+                plt.show()
 
-        if scatterplot:
-
-            ######## scatter plot after PCA ########
-            kns = np.nan_to_num(kns)
+            ######## scatter plot after PCA ########            
             pca = PCA(n_components=2)  # keep the first 2 components
             X_pca = pca.fit_transform(kns)
             X_pca = np.nan_to_num(X_pca)
             plot_components_2d(X_pca, y)
             plt.title('PCA')
-            plt.show()
 
+            if output_html:
+                html_str += plt2html(plt)
+                plt.close()
+            else:
+                plt.show()
+           
             ######## scatter plot after LDA ########
 
+            k_train, k_test, y_train, y_test = train_test_split(
+                    kns, y, test_size=0.3, random_state = 2, stratify=y)
+
             try:
-                kns = np.nan_to_num(kns)
                 lda = LinearDiscriminantAnalysis()
-                X_lda = lda.fit(kns, y).transform(kns)
+                X_lda = lda.fit(k_train, y_train).transform(k_test)
                 X_lda = np.nan_to_num(X_lda)
 
-                lda.score(kns, y)
+                lda.score(k_test, y_test)
                 if (X_lda.shape[1] == 1):
                     X_lda = np.hstack((X_lda, np.zeros((X_lda.shape[0], 1))))
-                plot_components_2d(X_lda, y)
+                plot_components_2d(X_lda, y_test)
                 # the coefficient of determination or R squared method is the proportion of the variance in the dependent variable that is predicted from the independent variable.
                 plt.title(
-                    'LDA (ACC = ' + str(np.round(lda.score(kns, y), 3)) + ')')
-                plt.show()
+                    'LDA (test acc = ' + str(np.round(lda.score(k_test, y_test), 3)) + ')')
+                
+                if output_html:
+                    html_str += plt2html(plt)
+                    plt.close()
+                else:
+                    plt.show()
+
             except Exception as e:
-                print('Exception : ', e)
+                if verbose:
+                    print('Exception : ', e)
                 # print('X_pls = ', X_pls)
                 # plt.title('PLS')
+                html_str += '<p>' + str(e) + '</p>'
 
             ######## scatter plot after PLS ########
 
-            try:
-                ''' # using CV
-                kns = np.nan_to_num(kns)
-                pls = PLSRegression(n_components=2, scale=False)
+            if len(set(y)) == 2: # skip PLS for multi-class cases
 
-                k_train, k_test, y_train, y_test = train_test_split(
-                    kns, y, test_size=0.3)
+                try:
+                    ''' # using CV
+                    kns = np.nan_to_num(kns)
+                    pls = PLSRegression(n_components=2, scale=False)
 
-                X_pls = pls.fit(k_train, y_train).transform(k_test)
-                X_pls = np.nan_to_num(X_pls)
+                    k_train, k_test, y_train, y_test = train_test_split(
+                        kns, y, test_size=0.3)
 
-                pls.score(k_test, y_test)
-                plot_components_2d(X_pls, y_test, legends=['C1', 'C2'])
-                # the coefficient of determination or R squared method is the proportion of the variance in the dependent variable that is predicted from the independent variable.
-                plt.title(
-                    'PLS (R2 = ' + str(np.round(pls.score(k_test, y_test), 3)) + ')')
-                plt.show()
-                '''
+                    X_pls = pls.fit(k_train, y_train).transform(k_test)
+                    X_pls = np.nan_to_num(X_pls)
 
-                kns = np.nan_to_num(kns)
-                pls = PLSRegression(n_components=2, scale=False)
-                X_pls = pls.fit(kns, y).transform(kns)
-                X_pls = np.nan_to_num(X_pls)
+                    pls.score(k_test, y_test)
+                    plot_components_2d(X_pls, y_test, legends=['C1', 'C2'])
+                    # the coefficient of determination or R squared method is the proportion of the variance in the dependent variable that is predicted from the independent variable.
+                    plt.title(
+                        'PLS (R2 = ' + str(np.round(pls.score(k_test, y_test), 3)) + ')')
+                    plt.show()
+                    '''
 
-                pls.score(kns, y)
-                plot_components_2d(X_pls, y)
-                # the coefficient of determination or R squared method is the proportion of the variance in the dependent variable that is predicted from the independent variable.
-                plt.title(
-                    'PLS (R2 = ' + str(np.round(pls.score(kns, y), 3)) + ')')
-                plt.show()
-            except Exception as e:
-                print('Exception : ', e)
-                # print('X_pls = ', X_pls)
-                # plt.title('PLS')
+                    kns = np.nan_to_num(kns)
+                    pls = PLSRegression(n_components=2, scale=False)
+                    X_pls = pls.fit(kns, y).transform(kns)
+                    X_pls = np.nan_to_num(X_pls)
 
-        ###### metrics ######
-        if metrics:
+                    pls.score(kns, y)
+                    plot_components_2d(X_pls, y)
+                    # the coefficient of determination or R squared method is the proportion of the variance in the dependent variable that is predicted from the independent variable.
+                    plt.title(
+                        'PLS (R2 = ' + str(np.round(pls.score(kns, y), 3)) + ')')
+                    if output_html:
+                        html_str += plt2html(plt)
+                        plt.close()
+                    else:
+                        plt.show()
+                except Exception as e:
+                    if verbose:
+                        print('Exception : ', e)
+                    # print('X_pls = ', X_pls)
+                    # plt.title('PLS')
+                    html_str += '<p>' + str(e) + '</p>'
+
+        ###### Classifiction after kernel transformation #######
+        if clfs == 'all' or len(clfs) > 0:
+            # result_html = '<h3>classification</h3>'
+            dic, result_html = run_multiclass_clfs(kns, y, clfs=clfs, show = False)
+            dic_test_accs[key] = dic
+            if output_html:
+                html_str += result_html
+            if plots:
+                IPython.display.display(IPython.display.HTML(result_html))
+
+        ###### cla metrics ######
+        if do_cla: # disable for now, keep just for reproduce historical results
             kns = np.nan_to_num(np.hstack((kns, np.array(y).reshape(-1, 1))),   # do nan filtering simultaneously for X and y
                                 nan=0, posinf=kns.max(), neginf=kns.min())
-            _, dic_metrics = get_metrics(kns[:, :-1], kns[:, -1].flatten())
-            dic_metrics['NMD'] = metric_nmd
+            _, dic_metrics = get_metrics(kns[:, :-1], kns[:, -1].flatten(), verbose = verbose)
+            # dic_metrics['NMD'] = metric_nmd
             all_dic_metrics[key] = dic_metrics
 
-    return all_dic_metrics
+    pickle.dump((KX, dic_test_accs, y), open('single_kernels.pkl', 'wb'))
 
+    if clfs == 'all' or len(clfs) > 0: # multi-kernel cases
+
+        # find top k kernels
+        best_KX = {}
+        best_top1s = {}
+        for k,v in dic_test_accs.items():
+            if k == 'no kernel':
+                continue
+            
+            best_top1 = 0
+            for kk, vv in v.items():
+                if best_top1 < vv[0]: # top-1 test acc
+                    best_top1 = vv[0]
+            best_top1s[k] = best_top1
+
+        keys = list(best_top1s.keys())
+        values = list(best_top1s.values())
+        sorted_value_index = np.argsort(values)[::-1]
+        sorted_dict = {keys[i]: values[i] for i in sorted_value_index}
+
+        print('Sorted single-kernel test accs: ', sorted_dict)
+
+        for k in sorted_dict: # sorted(best_top1s.items(),  key = lambda kv:(kv[1], kv[0]), reverse=True):
+            best_KX[k] = KX[k]
+            if multi_kernel_topk != -1 and len(best_KX) >= multi_kernel_topk:
+                break
+
+        for multi_kernel in multi_kernels:
+
+            # if multi_kernel == 1: # single kernel case, skip
+            #     continue
+
+            for idx, ks in enumerate(combinations(best_KX, multi_kernel)):
+                combined = np.zeros((len(y),0))
+                mk_title = ''
+                for iidx, k in enumerate(ks):
+                    if iidx == 0:
+                        mk_title += k
+                    else:
+                        mk_title += ' + ' + k
+                    combined = np.hstack((combined, KX[k]))
+                
+                dic, result_html = run_multiclass_clfs(combined, y, clfs=clfs, show = False)
+                result_html = '<h3>' + str(multi_kernel) + '-kernel ' + str(idx+1) + '. ' + mk_title + '</h3>' + result_html
+                dic_test_accs[mk_title] = dic
+                if output_html:
+                    html_str += result_html
+                if plots:
+                    IPython.display.display(IPython.display.HTML(result_html))
+
+    return KX, dic_test_accs, all_dic_metrics, html_str
+
+def visualize_kernel_result_dict(dic_test_accs):
+    '''
+    Visualize the result dictionary returned by classify_with_kernels()
+    '''
+
+    top1_accs = []
+    top3_accs = []
+    top5_accs = []
+    html_str = '<table>'
+    for k, v in dic_test_accs.items():
+        for kk, vv in v.items():
+            accs = [str(round(x,3)) for x in vv]
+            top1_accs.append(vv[0])
+            top3_accs.append(vv[1])
+            top5_accs.append(vv[2])
+            html_str += '<tr>' + '<td>' + k + '</td>'  + '<td>' + kk + '</td>' + '<td>' + accs[0] + '</td>' + '<td>' + accs[1] + '</td>' + '<td>' + accs[2] + '</td></tr>'
+
+    html_str += '<tr>' + '<td>best</td>' + '<td></td>' + '<td>' + \
+    str(round(np.max(top1_accs),3)) + '</td>' + '<td>' + \
+    str(round(np.max(top3_accs),3)) + '</td>' + '<td>' + \
+    str(round(np.max(top5_accs),3)) + '</td></tr>'
+
+    html_str += '</table>'
+    IPython.display.display(IPython.display.HTML(html_str))
 
 def visualize_metric_dicts(dics, plot=True):
     '''
     Example
     -------
-    dics = preview_kernels(X, y)
+    _, dics, _ = classify_with_kernels(X, y)
     html_str = generate_html_from_dicts(dics)
     display(HTML(html_str)) # use in jupyter notebook
     '''
@@ -356,17 +523,17 @@ def visualize_metric_dicts(dics, plot=True):
     row_names = []
     column_names = []
 
-    html_str = '<table><tr><th></th>'
+    html_str = '<table><tr><th>metric</th>'
 
     # use the 1st loop to get row and col names
     for kernel in dics:
         column_names.append(kernel)
-        html_str += '<th>' + kernel + '</th>'
+        html_str += '<th>' + str(kernel) + '</th>'
         for key in dics[kernel]:
             if key not in row_names:
                 row_names.append(key)
 
-    html_str += '<th>best kernel(s)</th>'
+    html_str += '<th>best</th><th>chart</th>'
     html_str += '</tr>'
 
     # use the 2nd loop to fill in data
@@ -401,21 +568,27 @@ def visualize_metric_dicts(dics, plot=True):
 
         best_kernel_names = str(np.array(column_names)[best_metric_idxs])
         html_str += '<td>' + best_kernel_names + '</td>'
-        html_str += '</tr>'
+        
+        if len(column_names) > 1: # only need to draw bar chart when more than two kernels are compared
 
-        if plot:
-            plt.figure(figsize=(20, 3))            
+            plt.figure(figsize=(20, 3))
             plt.bar(column_names, metrics, alpha=0.7,
-                    width=0.6, edgecolor='black', color='white', label = row)
+                    width=0.6, edgecolor='black', color='white', label=row)
             plt.bar(np.array(column_names)[best_metric_idxs], metrics[best_metric_idxs], alpha=0.7,
                     width=0.6, edgecolor='black', color='gold',
-                    label = "best kernels: " + best_kernel_names)
+                    label="best kernels: " + best_kernel_names)
             plt.xticks(rotation=40)
+            plt.title(row + "\nbest kernels: " + best_kernel_names +
+                    "\nbest value: " + str(best_metric_value))
             plt.legend()
-            plt.show()
 
-            print(row + "\nbest kernels: " + best_kernel_names +
-            "\nbest value: " + str(best_metric_value))
+            if plot:
+                plt.show()
+            else:
+                html_str += '<td style="min-width:400px">' + plt2html(plt) + '</td>'
+                plt.close()
+
+        html_str += '</tr>'
 
     html_str += '</table>'
     # display(HTML( html_str ))
@@ -439,8 +612,7 @@ def optimize_kernel_hparam(X, y, key, hparams=[], cmap=None):
         title = (kernel_fullnames[key] if key in kernel_fullnames else key) \
             + ('(' + format(h, '.2g') + ')')
         plt.imshow(kernel_dict[key](X, X, h), cmap=cmap)
-        metric_str = "NMD = %.3g" % nmd(X, y, lambda x, y: kernel_dict[key](x, y, h)) \
-            + "\tACC = %.3g" % acc(X, y, lambda x,
+        metric_str = "ACC = %.3g" % acc(X, y, lambda x,
                                    y: kernel_dict[key](x, y, h))
         plt.axis('off')
         plt.title(title + '\n' + kernel_formulas[key] + '\n' + metric_str)
@@ -473,8 +645,7 @@ def cosine_kernel_response_pattern(n=10, cmap='gray', logplot=False, embed_title
         ax[1].set_axis_off()
     plt.axis('off')
 
-    metric_str = "NMD = %.3g" % nmd(X, y, lambda x, y: kernel_dict['cosine'](x, y)) \
-        + "  ACC = %.3g" % acc(X, y, lambda x, y: kernel_dict['cosine'](x, y))
+    metric_str = "ACC = %.3g" % acc(X, y, lambda x, y: kernel_dict['cosine'](x, y))
 
     if embed_title:
         plt.title(kernel_fullnames['cosine'] + '\n' + kernel_formulas['cosine']
@@ -517,12 +688,13 @@ def time_cost_kernels(X, repeat=10, display=True):
         plt.show()
 
     dct_mu = {}
-    for k,v in zip(x, y.mean(axis = 1)):
+    for k, v in zip(x, y.mean(axis=1)):
         dct_mu[k] = v
 
     return dct_mu
 
-def plot_components_2d(X, y, labels=None, use_markers=False, ax=None, legends=None, tags=None):
+
+def plot_components_2d(X, y, labels=None, use_markers=False, legends=None, tags=None):
     '''
     Copied from qsi.vis.plot_components_2d, to avoid package dependency.
     '''
@@ -531,20 +703,26 @@ def plot_components_2d(X, y, labels=None, use_markers=False, ax=None, legends=No
             'ERROR: X MUST HAVE AT LEAST 2 FEATURES/COLUMNS! SKIPPING plot_components_2d().')
         return
 
-    # Gray shades can be given as a string encoding a float in the 0-1 range
-    colors = ['0.9', '0.1', 'red', 'blue', 'black',
-              'orange', 'green', 'cyan', 'purple', 'gray']
-    markers = ['o', 's', '^', 'D', 'H', 'o', 's', '^', 'D',
-               'H', 'o', 's', '^', 'D', 'H', 'o', 's', '^', 'D', 'H']
-
-    if ax is None:
-        _, ax = plt.subplots()
-
     if y is None or len(y) == 0:
         labels = [0]  # only one class
     if labels is None:
         labels = set(y)
 
+    _, ax = plt.subplots(figsize = (round(len(labels)/4.0) + 4, round(len(labels)/4.0) + 3))
+
+    # Gray shades can be given as a string encoding a float in the 0-1 range
+    colors = ['0.9', '0.1', 'red', 'blue', 'green',
+              'orange', 'cyan', 'purple', 'gray']
+    if len(labels) > len(colors):
+        for i in range(len(labels) - len(colors) + 5):
+            colors.append('C' + str(i))
+
+    markers = ['o', 's', '^', 'D', 'H', 'o', 's', '^', 'D',
+               'H', 'o', 's', '^', 'D', 'H', 'o', 's', '^', 'D', 'H']
+    if len(labels) > len(markers):
+        for i in range(len(labels) - len(markers) + 5):
+            markers.append(np.random.choice(markers))
+
     i = 0
 
     for label in labels:
